LOGGING = dict(
    version=1,
    disable_existing_loggers=True,
    objects={
        "queue": {
            "class":   "queue.Queue",
            "maxsize": 10000
        }
    },
    formatters={
        "simple":       {
            "format": "[%(asctime)s][%(levelname)s]: %(message)s"
        },
        "consoleDebug": {
            "class":  "flask_logify.formatters.RequestFormatter",
            "format": "[%(asctime)s]"
                      "[%(levelname)s]"
                      "[%(request_id)s]"
                      "[%(name)s:%(module)s.%(funcName)s:%(lineno)d]: "
                      "%(message)s"
        },
        "console":      {
            "class":  "flask_logify.formatters.RequestFormatter",
            "format": "[%(asctime)s][%(levelname)s][%(request_id)s]: %(message)s"
        },
        "syslog":       {
            "class":  "flask_logify.formatters.RequestFormatter",
            "format": f"%(ident)s%(message)s"
        }
    },
    handlers={
        "simple":       {
            "class":     "logging.StreamHandler",
            "formatter": "simple",
            "stream":    "ext://sys.stderr"
        },
        "console":      {
            "class":     "logging.StreamHandler",
            "formatter": "console",
            "stream":    "ext://sys.stderr"
        },
        "consoleDebug": {
            "class":     "logging.StreamHandler",
            "formatter": "consoleDebug",
            "stream":    "ext://sys.stderr"
        },
        "syslog":       {
            "class":     "flask_logify.handlers.FlaskSysLogHandler",
            "address":   ["localhost", 514],
            "formatter": "syslog",
            "facility":  "user"
        },
        "queueConsole": {
            "respect_handler_level": True,
            "class":                 "flask_logify.handlers.QueueHandler",
            "queue":                 "cfg://objects.queue",
            "handlers":              ["cfg://handlers.console"]
        },
        "queueSyslog":  {
            "respect_handler_level": True,
            "class":                 "flask_logify.handlers.QueueHandler",
            "queue":                 "cfg://objects.queue",
            "handlers":              ["cfg://handlers.syslog"]
        }
    },
    root={
        "level":    "WARN",
        "handlers": ["simple"]
    },
    loggers={
        "development":      {
            "level":     "DEBUG",
            "propagate": False,
            "handlers":  ["consoleDebug"]
        },
        "developmentQueue": {
            "level":     "DEBUG",
            "propagate": False,
            "handlers":  ["queueConsole"]
        },
        "production":       {
            "level":     "INFO",
            "propagate": False,
            "handlers":  ["console"]
        },
        "productionQueue":  {
            "level":     "INFO",
            "propagate": False,
            "handlers":  ["queueConsole"]
        },
        "flask-limiter":    {
            "level":     "DEBUG",
            "propagate": False,
            "handlers":  ["console"]
        },
        "gunicorn.error":   {
            "level":     "INFO",
            "propagate": False,
            "handlers":  ["queueSyslog"]
        },
        "gunicorn.access":  {
            "level":     "INFO",
            "propagate": False,
            "handlers":  ["queueSyslog"]
        }
    }
)

if __name__ == '__main__':  # pragma: no cover
    import json

    print(json.dumps(LOGGING))



import os
import sys

import jinja2
from werkzeug.middleware.lint import LintMiddleware
from werkzeug.middleware.profiler import ProfilerMiddleware

from flaskel import config, flaskel
from .converters import CONVERTERS
from .utils import misc, ObjectDict


class AppBuilder:
    """Flask app builder"""

    """
    default app name
    """
    app_name = config.FLASK_APP

    """
    default flask class
    """
    flask_class = flaskel.Flaskel

    """
    default secret key file name
    """
    secret_file = '.secret.key'

    """
    custom url converters
    """
    url_converters = CONVERTERS

    """
    custom app config module
    """
    conf_module = config

    def __init__(self, conf_module=None, extensions=None, converters=None,
                 blueprints=None, folders=None, middlewares=None, views=None,
                 after_request=None, before_request=None, callback=None, version=None, **options):
        """

        :param conf_module: python module file
        :param converters: custom url converter mapping
        :param extensions: custom extension appended to defaults
        :param blueprints: list of application's blueprints
        :param folders: list of custom jinja2's template folders
        :param middlewares: list of wsgi middleware
        :param views: list of views to register
        :param after_request: list of functions to register after request
        :param before_request: list of functions to register before request
        :param callback: a function called to patch app after all components registration
        :param options: passed to Flask class
        :param version: app version, may be used by extensions
        :return:
        """
        self._app = None
        self._version = version
        self._converters = converters or {}
        self._blueprints = blueprints or ()
        self._extensions = extensions or {}
        self._middlewares = middlewares or ()
        self._folders = folders or ()
        self._options = options or ()
        self._views = views or ()
        self._after_request = after_request or ()
        self._before_request = before_request or ()
        self._callback = callback or (lambda: None)
        self._conf_module = conf_module or self.conf_module

    def _generate_secret_key(self, secret_file, key_length):
        secret_key = misc.random_string(key_length)

        with open(secret_file, 'w') as f:
            f.write(secret_key)
            abs_file = os.path.abspath(secret_file)
            self._app.logger.warning("new secret key generated: take care of this file: %s", abs_file)
        return secret_key

    def _load_secret_key(self, secret_file):
        if not os.path.isfile(secret_file):
            return None

        with open(secret_file, 'r') as f:
            secret_key = f.read()
            self._app.logger.info("load secret key from: %s", secret_file)
        return secret_key

    def _set_secret_key(self):
        secret_key = None
        key_length = self._app.config.SECRET_KEY_MIN_LENGTH

        if not self._app.config.get('FLASK_ENV', '').lower().startswith('dev'):
            secret_file = self._app.config.SECRET_KEY
            if secret_file:
                secret_key = self._load_secret_key(secret_file) or secret_file
            else:
                secret_file = self.secret_file
                secret_key = self._load_secret_key(secret_file)
                secret_key = secret_key or self._generate_secret_key(secret_file, key_length)
        elif not self._app.config.get('SECRET_KEY'):
            self._app.logger.debug('set secret key in development mode')
            secret_key = 'fake_very_complex_string'

        self._app.config.SECRET_KEY = secret_key or self._app.config.SECRET_KEY
        if len(secret_key) < key_length:
            self._app.logger.warning("secret key length is less than: %s", key_length)

    def _register_extensions(self):
        with self._app.app_context():
            for name, e in self._extensions.items():
                try:
                    ext, opt = e[0], e[1] if len(e) > 1 else {}
                    if not ext:
                        raise TypeError("extension could not be None or empty")
                except (TypeError, IndexError) as exc:
                    self._app.logger.warning("Invalid extension '%s' configuration '%s': %s", name, e, exc)
                    continue

                ext.init_app(self._app, **opt)
                self._app.logger.debug("Registered extension '%s' with options: %s", name, opt)

        self._app.logger.debug("Dump flask extensions:")
        for k, v in self._app.extensions.items():
            self._app.logger.debug("Registered extension '%s': %s", k, v)

    def _register_blueprints(self):
        for b in self._blueprints:
            try:
                bp, opt = b[0], b[1] if len(b) > 1 else {}
                if not bp:
                    raise TypeError('blueprint could not be None or empty')
            except (TypeError, IndexError) as exc:
                self._app.logger.warning("invalid blueprint configuration '%s': %s", b, exc)
                continue

            self._app.register_blueprint(bp, **opt)
            self._app.logger.debug("Registered blueprint '%s' with options: %s", bp.name, opt)

    def _set_config(self, conf):
        self._app.config.from_object(self._conf_module)
        self._app.config.from_mapping(**(conf or {}))
        self._app.config.from_envvar('APP_CONFIG_FILE', silent=True)
        self._app.config = ObjectDict(**self._app.config)

    def _register_converters(self):
        self._app.url_map.converters.update({**self.url_converters, **self._converters})
        for k, v in self._app.url_map.converters.items():
            self._app.logger.debug("Registered converter: '%s' = %s", k, v.__name__)

    def _register_template_folders(self):
        loaders = [self._app.jinja_loader]

        for fsl in self._folders:
            loaders.append(jinja2.FileSystemLoader(fsl))

        if self._folders:
            self._app.jinja_loader = jinja2.ChoiceLoader(loaders)
            for f in self._folders:
                self._app.logger.debug("Registered template folder: '%s'", f)

    def _register_middlewares(self):
        for m in self._middlewares:
            kwargs = {}
            try:
                if isinstance(m, (list, tuple)):
                    m, kwargs = m[0], m[1] if len(m) > 1 else {}
                if not m:
                    raise TypeError('middleware could not be None or empty')
            except (TypeError, IndexError) as exc:
                self._app.logger.warning("invalid middleware configuration '%s': %s", m, exc)
                continue

            # WorkAround: in order to pass flask app to middleware without breaking chain
            if not (hasattr(m, 'flask_app') and m.flask_app):
                setattr(m, 'flask_app', self._app)

            self._app.wsgi_app = m(self._app.wsgi_app, **kwargs)
            self._app.logger.debug("Registered middleware: '%s'", m.__name__)

    def _register_views(self):
        def normalize(data):
            d = data + (None,) * (3 - len(data))
            if isinstance(d[1], dict):
                return d[0], d[2], d[1]
            return d[0], d[1], d[2] or {}

        for view in self._views:
            v, b, p = normalize(view)
            v.register(b or self._app, **p)
            mess = f"on blueprint '{b.name}'" if b else ""
            self._app.logger.debug("Registered view: '%s' %s with params: %s", v.__name__, mess, p)

    def _register_after_request(self):
        for f in self._after_request:
            self._app.after_request_funcs.setdefault(None, []).append(f)
            self._app.logger.debug("Registered function after request: '%s'", f.__name__)

    def _register_before_request(self):
        for f in self._before_request:
            self._app.before_request_funcs.setdefault(None, []).append(f)
            self._app.logger.debug("Registered function before request: '%s'", f.__name__)

    def _set_linter_and_profiler(self):
        if self._app.config.WSGI_WERKZEUG_LINT_ENABLED:
            self._app.wsgi_app = LintMiddleware(self._app.wsgi_app)
            self._app.logger.debug("Registered middleware: '%s'", LintMiddleware.__name__)

        if self._app.config.WSGI_WERKZEUG_PROFILER_ENABLED:
            file = self._app.config.WSGI_WERKZEUG_PROFILER_FILE
            self._app.wsgi_app = ProfilerMiddleware(
                self._app.wsgi_app,
                stream=open(file, 'w') if file else sys.stdout,
                restrictions=self._app.config.WSGI_WERKZEUG_PROFILER_RESTRICTION
            )
            self._app.logger.debug("Registered middleware: '%s'", ProfilerMiddleware.__name__)

    def _dump_urls(self):
        class DumpUrls:
            def __init__(self, rules):
                self._rules = rules

            def __str__(self):
                output = []
                for rule in self._rules:
                    methods = ','.join(rule.methods)
                    output.append("{:30s} {:40s} {}".format(rule.endpoint, methods, rule))
                return "\n".join(sorted(output))

        self._app.logger.debug("Registered routes:\n%s", DumpUrls(self._app.url_map.iter_rules()))

    def _patch_app(self):
        if self._app.debug:
            self._set_linter_and_profiler()
            self._dump_urls()

        with self._app.app_context():
            sqlalchemy = self._app.extensions.get('sqlalchemy')
            if sqlalchemy is not None:
                sqlalchemy.db.create_all()

            self._callback()

    def create(self, conf=None):
        self._app = self.flask_class(self.app_name, **self._options)
        self._app.version = self._version

        self._set_config(conf)
        self._set_secret_key()

        self._register_extensions()
        self._register_middlewares()
        self._register_template_folders()
        self._register_converters()
        self._register_views()
        self._register_blueprints()
        self._register_after_request()
        self._register_before_request()

        self._patch_app()

    def get_or_create(self, conf=None):
        if self._app is None:
            self.create(conf)

        return self._app



from .batch import FlaskelHTTPBatch, HTTPBatch
from .client import (
    all_errors, FlaskelHttp, FlaskelJsonRPC, HTTPBase, HTTPClient,
    HTTPStatusError, HTTPTokenAuth, NetworkError
)
from .httpdumper import BaseHTTPDumper, FlaskelHTTPDumper, LazyHTTPDumper
from .useragent import UserAgent



# noinspection PyUnresolvedReferences
from flask import current_app as cap

from .builder import AppBuilder
from .flaskel import Flaskel, Request, Response
from .http import exceptions as HttpError, httpcode
from .standalone import Server
from .tester import TestClient
from .utils import ConfigProxy, ExtProxy, misc, ObjectDict, uuid, webargs, yaml
from .version import *



from flask import Blueprint

from flaskel.ext import error_handler

bp_web = Blueprint(
    'web', __name__,
    template_folder="templates",
    static_folder="static"
)

error_handler.web_register(bp_web)



from .batch import AsyncBatchExecutor, BatchExecutor, DaemonThread, Thread, ThreadBatchExecutor
from .datastruct import ConfigProxy, ExtProxy, HashableDict, IntEnum, ObjectDict
from .datetime import Day, from_iso_format, Millis, Minutes, Seconds, to_iso_format
from .schemas import JSONSchema, PayloadValidator, SCHEMAS



import flask
from flaskel.flaskel import cap, httpcode
from flaskel.utils import PayloadValidator, webargs
from sqlalchemy.exc import IntegrityError, SQLAlchemyError

from .base import Resource


class CatalogResource(Resource):
    methods_collection = ['GET']
    methods_resource = ['GET']
    methods_subresource = ['GET']

    def __init__(self, model):
        """

        :param model: SqlAlchemy model class
        """
        self._model = model

    def on_get(self, res_id, model=None, *args, **kwargs):
        """
        Get resource info

        :param res_id: resource identifier (primary key value)
        :param model: alternative SqlAlchemy model class
        :param kwargs: extra query filters
        :return:
        """
        model = model or self._model
        return model.get_one(id=res_id, **kwargs)

    def on_collection(self, params=None, model=None, *args, **kwargs):
        """
        Resource collection paginated and sorted

        :param params: parameters (usually from query string)
        :param model: alternative SqlAlchemy model class
        :param kwargs: extra query filters
        :return:
        """
        if params is None:
            params = webargs.paginate()

        model = model or self._model
        max_size = cap.config.MAX_PAGE_SIZE
        page = params.get('page')
        size = params.get('page_size')
        size = max(size, max_size or 0) if size else max_size
        order_by = getattr(model, 'order_by', None)

        return self.response_paginated(
            model.get_list(
                to_dict=False, order_by=order_by, page=page, page_size=size,
                max_per_page=max_size, params=params, **kwargs
            ),
            restricted=not params.get('related', False)
        )

    def response_paginated(self, res, **kwargs):
        """
        Prepare the paginated response for resource collection

        :param res: list of sqlalchemy models
        :return:
        """
        if type(res) is list:
            return [r.to_dict(**kwargs) for r in res]

        return (
            [r.to_dict(**kwargs) for r in res.items],
            httpcode.PARTIAL_CONTENT if res.has_next else httpcode.SUCCESS,
            self.pagination_headers(res)
        )

    @staticmethod
    def pagination_headers(data):
        return {
            'X-Pagination-Count':     data.total,
            'X-Pagination-Page':      data.page,
            'X-Pagination-Num-Pages': data.pages,
            'X-Pagination-Page-Size': data.per_page,
        }


class Restful(CatalogResource):
    post_schema = None
    put_schema = None
    validator = PayloadValidator
    methods_collection = ['GET', 'POST']
    methods_subresource = ['GET', 'POST']
    methods_resource = ['GET', 'PUT', 'DELETE']

    def __init__(self, session, model):
        """

        :param session: sqlalchemy session instance
        :param model: sqlalchemy model
        """
        super().__init__(model)
        self._session = session

    def validate(self, schema):
        """

        :param schema: schema compatible with self validator
        :return:
        """
        schema = schema() if callable(schema) else schema
        return self.validator.validate(schema)

    def create_resource(self, data):
        """

        :param data: dictionary data that represents the resource
        :return: sqlalchemy model instance
        """
        return self._model(**data)

    # noinspection PyMethodMayBeStatic
    def update_resource(self, resource, data):
        """

        :param resource: sqlalchemy model instance
        :param data: dictionary data that represents the resource
        :return:
        """
        resource.update(data)
        return resource

    def _create(self, res):
        try:
            self._session.add(res)
            self._session.commit()
        except IntegrityError:
            self._session.rollback()
            flask.abort(httpcode.CONFLICT)
        except SQLAlchemyError as exc:
            cap.logger.exception(exc)
            self._session.rollback()
            flask.abort(httpcode.INTERNAL_SERVER_ERROR)

    def on_post(self, *args, **kwargs):
        """

        :return:
        """
        payload = self.validate(self.post_schema)
        res = self.create_resource(payload)
        self._create(res)
        return res.to_dict(), httpcode.CREATED

    def on_delete(self, res_id, *args, **kwargs):
        """

        :param res_id: resource identifier (primary key value)
        """
        res = self._model.query.get_or_404(res_id)

        try:
            self._session.delete(res)
            self._session.commit()
        except SQLAlchemyError as exc:
            cap.logger.exception(exc)
            self._session.rollback()
            flask.abort(httpcode.INTERNAL_SERVER_ERROR)

    def _update(self, res):
        try:
            self._session.merge(res)
            self._session.commit()
        except SQLAlchemyError as exc:
            cap.logger.exception(exc)
            self._session.rollback()
            flask.abort(httpcode.INTERNAL_SERVER_ERROR)

    def on_put(self, res_id, *args, **kwargs):
        """

        :param res_id: resource identifier (primary key value)
        :return:
        """
        payload = self.validate(self.put_schema)
        res = self._model.query.get_or_404(res_id)
        res = self.update_resource(res, payload)
        self._update(res)
        return res.to_dict()



# example auth blueprint
#
from flask import Blueprint

from flaskel.ext import cors, error_handler

bp_auth = Blueprint('auth', __name__, subdomain='api', url_prefix='/auth')

cors.init_app(bp_auth)
error_handler.api_register(bp_auth)



import os

from decouple import Choices, config, Csv

from flaskel.utils import logger, yaml

DEBUG = config('DEBUG', default=False, cast=bool)
TESTING = config('TESTING', default=DEBUG, cast=bool)
APP_NAME = config('APP_NAME', default='flaskel')
APP_HOST = config('APP_HOST', default='127.0.0.1')
APP_PORT = config('APP_PORT', default=5000, cast=int)
FLASK_APP = config('FLASK_APP', default='app:app')
SERVER_NAME = config('SERVER_NAME', default=f'{APP_HOST}:{APP_PORT}')

FLASK_ENV = config(
    'FLASK_ENV',
    default='development' if DEBUG else 'production',
    cast=Choices(['development', 'production'])
)

LOCALE = config('LOCALE', default='en_EN.utf8')
TEMPLATES_AUTO_RELOAD = config('TEMPLATES_AUTO_RELOAD', default=DEBUG, cast=bool)
EXPLAIN_TEMPLATE_LOADING = config('EXPLAIN_TEMPLATE_LOADING', default=False, cast=bool)
APIDOCS_ENABLED = config('APIDOCS_ENABLED', default=True, cast=bool)
CONF_PATH = config('CONF_PATH', default=os.path.join('flaskel', 'scripts', 'skeleton', 'res'))

SQLALCHEMY_DATABASE_URI = config('DATABASE_URL', default='sqlite:///db.sqlite')
REDIS_URL = config('REDIS_URL', default='redis://127.0.0.1:6379')
REDIS_OPTS = {
    'socket_connect_timeout': config('REDIS_CONN_TIMEOUT', default=0.05, cast=float)
}

BASIC_AUTH_USERNAME = config('BASIC_AUTH_USERNAME', default='admin')
BASIC_AUTH_PASSWORD = config('BASIC_AUTH_PASSWORD', default='admin')

MAIL_DEBUG = config('MAIL_DEBUG', default=DEBUG, cast=bool)
MAIL_SERVER = config('MAIL_SERVER', default='sendria.local')
MAIL_PORT = config('MAIL_SERVER', default=62000, cast=int)
ADMIN_EMAIL = config('ADMIN_EMAIL', default='admin')
ADMIN_PASSWORD = config('ADMIN_PASSWORD', default='admin')
MAIL_DEFAULT_SENDER = config('MAIL_DEFAULT_SENDER', default='admin@mail.com')
MAIL_DEFAULT_RECEIVER = config('MAIL_DEFAULT_RECEIVER', default='admin@mail.com')

PREFERRED_URL_SCHEME = config(
    'PREFERRED_URL_SCHEME',
    default='http' if FLASK_ENV == 'development' else 'https'
)

IPBAN_COUNT = config('IPBAN_COUNT', default=5, cast=int)
IPBAN_SECONDS = config('IPBAN_SECONDS', default=3600, cast=int)

LOG_BUILDER = config('LOG_BUILDER', default='text')
LOG_APP_NAME = config('LOG_APP_NAME', default=APP_NAME)
LOG_LOGGER_NAME = config('LOG_LOGGER_NAME', default=FLASK_ENV)
LOG_REQ_HEADERS = config('LOG_REQ_HEADERS', default='', cast=Csv())
LOG_RESP_HEADERS = config('LOG_RESP_HEADERS', default='', cast=Csv())
LOG_REQ_SKIP_DUMP = config('LOG_REQ_SKIP_DUMP', default=not TESTING, cast=bool)
LOG_RESP_SKIP_DUMP = config('LOG_RESP_SKIP_DUMP', default=not TESTING, cast=bool)
LOG_REQ_FORMAT = 'INCOMING REQUEST: {address} {method} {scheme} {path}\n{headers}\n{body}'
LOG_RESP_FORMAT = 'OUTGOING RESPONSE for {address} at {path}: STATUS {status}\n{headers}\n{body}'

CF_STRICT_ACCESS = config('CF_STRICT_ACCESS', default=False, cast=bool)
VERSION_STORE_MAX = config('VERSION_STORE_MAX', default=6, cast=int)
VERSION_CACHE_EXPIRE = config('VERSION_CACHE_EXPIRE', default=60, cast=int)

HTTP_PROTECT_BODY = config('HTTP_PROTECT_BODY', default=False, cast=bool)
HTTP_DUMP_BODY = [
    config('HTTP_DUMP_REQ_BODY', default=False, cast=bool),
    config('HTTP_DUMP_RESP_BODY', default=False, cast=bool),
]

USE_X_SENDFILE = config('USE_X_SENDFILE', default=not DEBUG, cast=bool)
ENABLE_ACCEL = config('ENABLE_ACCEL', default=True, cast=bool)
ACCEL_BUFFERING = True
ACCEL_CHARSET = 'utf-8'
ACCEL_LIMIT_RATE = 'off'

WSGI_WERKZEUG_PROFILER_FILE = 'profiler.txt'
WSGI_WERKZEUG_PROFILER_RESTRICTION = (0.1,)
WSGI_WERKZEUG_LINT_ENABLED = config('WSGI_WERKZEUG_LINT_ENABLED', default=TESTING, cast=bool)
WSGI_WERKZEUG_PROFILER_ENABLED = config('WSGI_WERKZEUG_PROFILER_ENABLED', default=TESTING, cast=bool)

SQLALCHEMY_TRACK_MODIFICATIONS = False
SQLALCHEMY_ECHO = config('SQLALCHEMY_ECHO', default=TESTING, cast=bool)

REQUEST_ID_HEADER = 'X-Request-ID'
RATELIMIT_ENABLED = config('RATELIMIT_ENABLED', default=not DEBUG, cast=bool)
RATELIMIT_HEADERS_ENABLED = config('RATELIMIT_HEADERS_ENABLED', default=True, cast=bool)
RATELIMIT_IN_MEMORY_FALLBACK_ENABLED = config('RATELIMIT_IN_MEMORY_FALLBACK_ENABLED', default=True, cast=bool)

DATABASE_URL = SQLALCHEMY_DATABASE_URI
ERROR_PAGE = 'core/error.html'
SECRET_KEY_MIN_LENGTH = 256
RB_DEFAULT_ACCEPTABLE_MIMETYPES = [
    'application/json',
    'application/xml',
]

JWT_DEFAULT_SCOPE = None
JWT_DEFAULT_TOKEN_TYPE = 'bearer'
PRETTY_DATE = '%d %B %Y %I:%M %p'
DATE_ISO_FORMAT = '%Y-%m-%dT%H:%M:%S'
SEND_FILE_MAX_AGE_DEFAULT = 86400

RATELIMIT_STORAGE_URL = REDIS_URL
RATELIMIT_KEY_PREFIX = APP_NAME
RATELIMIT_STORAGE_OPTIONS = {
    'socket_timeout':         REDIS_OPTS['socket_connect_timeout'],
    'socket_connect_timeout': REDIS_OPTS['socket_connect_timeout']
}

LIMITER = {
    'FAIL':         '1/second',
    'FAST':         '30/minute',
    'MEDIUM':       '20/minute',
    'SLOW':         '10/minute',
    'BYPASS_KEY':   'X-Limiter-Bypass',
    'BYPASS_VALUE': 'bypass-rate-limit',
}

SCHEDULER_AUTO_START = config('SCHEDULER_AUTO_START', default=True, cast=bool)
SCHEDULER_API_ENABLED = config('SCHEDULER_API_ENABLED', default=False, cast=bool)
SCHEDULER_JOBSTORES = {
    'default': {
        'class':     'apscheduler.jobstores.sqlalchemy:SQLAlchemyJobStore',
        'tablename': 'scheduler',
        'url':       DATABASE_URL
    }
}
SCHEDULER_EXECUTORS = {
    'default': {
        'type':        'threadpool',
        'max_workers': 20
    }
}
SCHEDULER_JOB_DEFAULTS = {
    'coalesce':      False,
    'max_instances': 10
}

CACHE_REDIS_URL = REDIS_URL
CACHE_DEFAULT_TIMEOUT = 3600
CACHE_TYPE = 'flask_caching.backends.redis'
CACHE_KEY_PREFIX = config('CACHE_KEY_PREFIX', default=APP_NAME)
CACHE_OPTIONS = {
    'socket_timeout':         REDIS_OPTS['socket_connect_timeout'],
    'socket_connect_timeout': REDIS_OPTS['socket_connect_timeout']
}

APISPEC = yaml.load_optional_yaml_file(os.path.join(CONF_PATH, 'swagger.yaml'))
SCHEMAS = yaml.load_optional_yaml_file(os.path.join(CONF_PATH, 'schemas.yaml'))
SCHEDULER_JOBS = yaml.load_optional_yaml_file(os.path.join(CONF_PATH, 'scheduler.yaml'))
IPBAN_NUISANCES = yaml.load_optional_yaml_file(os.path.join(CONF_PATH, 'nuisances.yaml'))
LOGGING = yaml.load_optional_yaml_file(os.path.join(CONF_PATH, 'log.yaml'), default=logger.LOGGING)



import functools
from datetime import timedelta

import flask_jwt_extended as jwt
from flask_httpauth import HTTPBasicAuth

from flaskel.flaskel import cap, httpcode
from flaskel.utils.datastruct import ObjectDict
from .sqlalchemy import db
from .sqlalchemy.mixins import StandardMixin

jwtm = jwt.JWTManager()
basic_auth = HTTPBasicAuth()


@basic_auth.verify_password
def simple_basic_auth(username, password):
    if username == cap.config.BASIC_AUTH_USERNAME \
            and password == cap.config.BASIC_AUTH_PASSWORD:
        return dict(username=username, password=password)


@jwtm.invalid_token_loader
def invalid_token_loader(mess):
    return dict(message=mess), httpcode.UNAUTHORIZED  # pragma: no cover


class RevokedTokenMixin(StandardMixin):
    jti = db.Column(db.String(120), nullable=False, unique=True)

    def __repr__(self):  # pragma: no cover
        return f"<RevokedToken: {self.id} - {self.jti}>"

    @classmethod
    def is_block_listed(cls, jti):  # pragma: no cover
        # noinspection PyUnresolvedReferences
        return bool(cls.query.filter_by(jti=jti).first())


class BaseTokenHandler:
    def __init__(self, blocklist_loader=None):
        """

        :param blocklist_loader:
        """
        jwtm.token_in_blocklist_loader(blocklist_loader)

    def check_token_block_listed(self, jwt_headers, jwt_data):
        raise NotImplementedError()  # pragma: no cover

    def revoke(self, token=None):
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    def check_permission(cls, perm):
        """

        :param perm:
        """

    @classmethod
    def auth_required(cls, perm=None, **kwargs):
        def wrapper(f):
            @functools.wraps(f)
            @jwt.jwt_required(**kwargs)
            def check_optional_permission(*args, **kw):
                if perm is not None:
                    cls.check_permission(perm)
                return f(*args, **kw)

            return check_optional_permission

        return wrapper

    @classmethod
    def identity(cls):
        identity = jwt.get_jwt_identity()
        if type(identity) is dict:
            identity = ObjectDict(**identity)
        return identity

    @classmethod
    def get_raw(cls):
        return ObjectDict(**jwt.get_jwt())

    @classmethod
    def decode(cls, token):
        return ObjectDict(**jwt.decode_token(token))

    def get_access(self, identity=None, expires=None):
        """

        :param identity:
        :param expires:
        :return:
        """
        return jwt.create_access_token(
            identity=identity or self.identity(),
            expires_delta=timedelta(expires) if expires else None
        )

    def get_refresh(self, identity=None, expires=None):
        """

        :param identity:
        :param expires:
        :return:
        """
        return jwt.create_refresh_token(
            identity=identity or self.identity(),
            expires_delta=timedelta(expires) if expires else None
        )

    def refresh(self, expires=None):
        """

        :param expires: in seconds
        :return:
        """
        access_token = self.get_access(expires=expires)
        decoded = self.decode(access_token)
        return ObjectDict(
            access_token=access_token,
            expires_in=decoded.exp,
            issued_at=decoded.iat,
            token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
            scope=cap.config.JWT_DEFAULT_SCOPE
        )

    def create(self, identity, refresh=True, expires_access=None, expires_refresh=None, scope=None):
        """

        :param identity: user identifier, generally the username
        :param refresh: enable refresh token
        :param expires_access: in seconds
        :param expires_refresh: in seconds
        :param scope:
        :return:
        """
        access_token = self.get_access(identity=identity, expires=expires_access)
        decoded = self.decode(access_token)
        resp = ObjectDict(
            access_token=access_token,
            expires_in=decoded.exp,
            issued_at=decoded.iat,
            token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
            scope=scope or cap.config.JWT_DEFAULT_SCOPE
        )

        if refresh:
            resp.refresh_token = self.get_refresh(identity=identity, expires=expires_refresh)

        return resp

    def dump(self, token_type=None, scope=None):
        """

        :param token_type:
        :param scope:
        :return:
        """
        return ObjectDict(
            token_type=token_type or cap.config.JWT_DEFAULT_TOKEN_TYPE,
            scope=scope or cap.config.JWT_DEFAULT_SCOPE,
            **self.get_raw()
        )


class DBTokenHandler(BaseTokenHandler):
    def __init__(self, model, session, blocklist_loader=None):
        """

        :param model:
        """
        self.model = model
        self.session = session
        super().__init__(blocklist_loader or self.check_token_block_listed)

    def check_token_block_listed(self, jwt_headers, jwt_data):
        """

        :param jwt_headers:
        :param jwt_data:
        :return:
        """
        return self.model.is_block_listed(jwt_data['jti'])

    def revoke(self, token=None):
        """

        :param token:
        """
        token = self.decode(token) if token else self.get_raw()
        if token.jti:
            self.session.add(self.model(jti=token.jti))
            self.session.commit()


class RedisTokenHandler(BaseTokenHandler):
    entry_value = 'true'
    key_prefix = 'token_revoked::'

    def __init__(self, redis, blocklist_loader=None):
        """

        :param redis:
        :param blocklist_loader:
        """
        self.redis = redis
        super().__init__(blocklist_loader or self.check_token_block_listed)

    def check_token_block_listed(self, jwt_headers, jwt_data):
        """

        :param jwt_headers:
        :param jwt_data:
        :return:
        """
        entry = self.redis.get(f"{self.key_prefix}{jwt_data['jti']}")
        if not entry:
            return False

        return entry == self.entry_value or entry == self.entry_value.encode()

    def revoke(self, token=None):
        """

        :param token:
        """
        token = self.decode(token) if token else self.get_raw()
        self.redis.set(f"{self.key_prefix}{token.jti}", self.entry_value)



from flaskel.extra.mobile_support import MobileLoggerView, MobileReleaseView
from flaskel.views import apidoc, rpc

bp_api = None  # import blueprint here
rpc_service = object  # import rpc_service here
rpc.JSONRPCView.load_from_object(rpc_service())

VIEWS = (
    (rpc.JSONRPCView, dict(
        name=rpc.JSONRPCView.default_view_name,
        url=rpc.JSONRPCView.default_url
    )),
    (apidoc.ApiDocTemplate, bp_api, dict(
        name=apidoc.ApiDocTemplate.default_view_name,
        urls=apidoc.ApiDocTemplate.default_urls
    )),
    (apidoc.ApiSpecTemplate, bp_api, dict(
        name=apidoc.ApiSpecTemplate.default_view_name,
        urls=apidoc.ApiSpecTemplate.default_urls
    )),
    (MobileReleaseView, bp_api, dict(
        name=MobileReleaseView.default_view_name,
        urls=MobileReleaseView.default_urls,
    )),
    (MobileLoggerView, bp_api, dict(
        name=MobileLoggerView.default_view_name,
        urls=MobileLoggerView.default_urls,
    )),
)



from functools import partial

import flask
from webargs import fields, flaskparser

from flaskel import config, httpcode

parser = flaskparser.FlaskParser()
query = partial(parser.use_args, location="query")
payload = partial(parser.use_args, location="json")


class Field:
    integer = partial(fields.Integer)
    string = partial(fields.String)
    decimal = partial(fields.Decimal)
    boolean = partial(fields.Boolean, truthy={'true', 1}, falsy={'false', 0})
    positive = partial(integer, validate=lambda x: x > 0)
    not_negative = partial(integer, validate=lambda x: x >= 0)
    not_positive = partial(integer, validate=lambda x: x <= 0)
    negative = partial(integer, validate=lambda x: x < 0)
    isodate = partial(fields.DateTime, format=config.DATE_ISO_FORMAT)
    list_of = partial(fields.DelimitedList, cls_or_instance=string(), delimiter='+')


class OptField:
    integer = partial(Field.integer, missing=None)
    string = partial(Field.string, missing=None)
    decimal = partial(Field.decimal, missing=None)
    boolean = partial(Field.boolean, missing=False)
    positive = partial(Field.positive, missing=None)
    not_negative = partial(Field.not_negative, missing=None)
    not_positive = partial(Field.not_positive, missing=None)
    negative = partial(Field.negative, missing=None)
    isodate = partial(Field.isodate, missing=None)
    list_of = partial(Field.list_of, missing=())


class ReqField:
    integer = partial(Field.integer, required=True)
    string = partial(Field.string, required=True)
    decimal = partial(Field.decimal, required=True)
    boolean = partial(Field.boolean, required=True)
    positive = partial(Field.positive, required=True)
    not_negative = partial(Field.not_negative, required=True)
    not_positive = partial(Field.not_positive, required=True)
    negative = partial(Field.negative, required=True)
    isodate = partial(Field.isodate, required=True)
    list_of = partial(Field.list_of, required=True)


# noinspection PyUnusedLocal
@parser.error_handler
def handle_error(error, *args, **kwargs):
    flask.abort(httpcode.BAD_REQUEST, response=error.messages)


query_paginate = partial(query, dict(
    page=OptField.positive(),
    page_size=OptField.positive(),
    related=OptField.boolean(missing=False),
))


def paginate(f=None):
    @query_paginate()
    def _route(*args, **kwargs):
        params = args[-1]
        if f is None:
            return params

        return f(*args[:-1], params=params, **kwargs)

    return _route if f else _route()



from gevent.pywsgi import WSGIServer

from .base import BaseApplication


class WSGIGevent(BaseApplication, WSGIServer):
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        BaseApplication.__init__(self, app, options)
        opts = dict(
            spawn=options.get('spawn') or 'default',
            backlog=options.get('backlog')
        )
        WSGIServer.__init__(self, self._bind, self.application, **opts)

    def run(self):
        """

        """
        self.serve_forever()



import flask

from flaskel.flaskel import cap, httpcode
from .base import BaseView


class ApiDocTemplate(BaseView):
    apispec_view = 'api.apispec'
    default_view_name = 'apidocs'
    default_urls = ['/apidocs']

    def dispatch_request(self):
        if not cap.config.APIDOCS_ENABLED:
            flask.abort(httpcode.NOT_FOUND)  # pragma: no cover

        rapidoc_version = cap.config.RAPIDOC_VERSION or '8.4.3'
        rapidoc_theme = cap.config.RAPIDOC_THEME or 'dark'
        page_title = f"{cap.config.APP_NAME} - API DOCS"

        spec_url = flask.url_for(
            self.apispec_view,
            _external=True,
            _scheme=cap.config.PREFERRED_URL_SCHEME
        )

        return f"""<!doctype html>
<html>
    <head>
        <meta charset="utf-8"/>
        <meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1"/>
        <meta name="viewport" content="width=device-width,initial-scale=1,shrink-to-fit=no"/>
        <title>{page_title}</title>
        <script type="module"
            src="https://unpkg.com/rapidoc@{rapidoc_version}/dist/rapidoc-min.js">
        </script>
    </head>
    <body>
        <rapi-doc theme="{rapidoc_theme}" spec-url="{spec_url}"></rapi-doc>
    </body>
</html>"""


class ApiSpecTemplate(BaseView):
    default_view_name = 'apispec'
    default_urls = ['/apidoc.json']

    def __init__(self, version='1.0.0', context_path=''):
        """

        :param version:
        :param context_path:
        """
        self._api_version = version
        self._context_path = f"{flask.request.environ['SCRIPT_NAME']}/{context_path}"

    def dispatch_request(self):
        if not cap.config.APIDOCS_ENABLED:
            flask.abort(httpcode.NOT_FOUND)  # pragma: no cover

        return self.set_variables(cap.config.APISPEC)

    def set_variables(self, apispec):
        """

        :param apispec:
        :return:
        """
        try:
            apispec.info.version = self._api_version
            variables = apispec.servers[0].variables
            variables['context']['default'] = self._context_path
            scheme = cap.config.PREFERRED_URL_SCHEME or 'http'
            variables['host']['default'] = f"{scheme}://{cap.config.SERVER_NAME}"
        except (AttributeError, IndexError, KeyError, TypeError) as exc:  # pragma: no cover
            cap.logger.exception(exc)
        return apispec
